#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import abc

from typing import Callable, Optional, Type, TypeVar

import attrs
import click

from pyspark.sql import SparkSession
from typing_extensions import ParamSpec

from pramen_py.app import env
from pramen_py.models import RunTransformer, TransformationConfig
from pramen_py.transformation.transformation_base import (
    T_EXTRA_OPTIONS,
    Transformation,
)
from pramen_py.utils import get_or_create_spark_session


T_TRANSFORMATION = TypeVar("T_TRANSFORMATION", bound=Transformation)
CLI_CALLBACK = ParamSpec("CLI_CALLBACK")


def get_current_transformer_cfg(
    T: Type[T_TRANSFORMATION],
    config: TransformationConfig,
) -> RunTransformer:
    """Find corresponding transformer in the config.

    In case if the current transformer is not defined - raise an error.
    """
    try:
        transformer_to_run = next(
            filter(
                lambda t: t.name == T.__name__,
                config.run_transformers,
            )
        )
    except StopIteration:
        raise ValueError(
            f"{T.__name__} transformer missed "
            f"in the configuration. Currently the following transformers "
            f"are available in the configuration:\n"
            f"{chr(10).join(map(str,config.run_transformers))}"
        )
    else:
        return transformer_to_run


@attrs.define(auto_attribs=True, slots=True)
class Runner(metaclass=abc.ABCMeta):
    """Base Runner interface to manage Transformation run.

    Runner's implementations provides cli to control Transformations.
    Runners should be registered in pramen_py.app.cli.RunGroup.__init__
    after super().__init__(*args, **kwargs).

    There could be multiple runners registered in order to provide
    more control over the Transformations. For example the
    runner which executes all transformations asynchronously.
    """

    cli: click.Group = attrs.field()
    spark: SparkSession = attrs.field(default=None)

    @abc.abstractmethod
    def activate(self) -> None:
        """Runner entrypoint.

        Register cli interface, prepare resources etc.
        """
        ...

    @abc.abstractmethod
    def create_cli_cmd_for_transformations(self) -> None:
        """Setup cli command and interface for all available transformations.

        Use discover_transformations function to find all Transformation's
        subclasses.
        """
        ...

    @abc.abstractmethod
    def create_cli_cmd_callback(
        self,
        T: Type[T_TRANSFORMATION],
    ) -> Callable[CLI_CALLBACK, None]:
        """Define a transformation run callback.

        This is a command to be executed on a cli invocation.

        Since Transformation.run are defined as async function, use
        coro util function to prepare the event loop and schedule running
        the target using it. For example:

        >>> def create_cli_cmd_callback(
        >>>     self,
        >>>     T: Type[T_TRANSFORMATION],
        >>> ) -> Callable[CLI_CALLBACK, None]:
        >>>     async def t_run_wrapper(
        >>>         ctx: click.Context,
        >>>         config: TransformationConfig,
        >>>         info_date: Optional[str],
        >>>         **kwargs: T_EXTRA_OPTIONS,
        >>>     ) -> None:
        >>>         number_of_trials = env.int(
        >>>             "PRAMENPY_MAX_RETRIES_EXECUTE_TRANSFORMATION",
        >>>             1,
        >>>         )
        >>>         with logger.contextualize(
        >>>             transformer=f"{T.__name__}@{int(random() * 10000)}"
        >>>         ):
        >>>             await run_and_retry(
        >>>                 partial(
        >>>                     self.process_transformer,
        >>>                     T=T,
        >>>                     config=config,
        >>>                     info_date=info_date,
        >>>                     kwargs=kwargs,
        >>>                 ),
        >>>                 number_of_trials=number_of_trials,
        >>>                 exception=Py4JJavaError,
        >>>                 retry_hook=self.process_transformer_retry_hook,
        >>>             )

        >>>     return click.pass_context(coro(t_run_wrapper))
        """
        ...

    @abc.abstractmethod
    async def process_transformer(
        self,
        T: Type[T_TRANSFORMATION],
        config: TransformationConfig,
        info_date: Optional[str],
        **kwargs: T_EXTRA_OPTIONS,
    ) -> None:
        """Prepare and run the transformation.

        :param info_date: if specified should have precedence over the config
            value (use overwrite_info_dates utils function to update the
            config)

        IMPORTANT: the implementation should call the super class default
            implementation first, i.e.:

        >>> async def process_transformer(
        >>>     self,
        >>>     T: Type[T_TRANSFORMATION],
        >>>     config: TransformationConfig,
        >>>     info_date: Optional[str],
        >>>     **kwargs: T_EXTRA_OPTIONS,
        >>> ) -> None:
        >>>     await super().process_transformer(
        >>>         T,
        >>>         config,
        >>>         info_date,
        >>>         **kwargs,
        >>>     )

        building spark session at this stage makes sense, because
        this prepares the spark session (expensive) only in case of processing
        the tranformation.
        """

        self.spark = self.spark or get_or_create_spark_session(
            env,
            spark_config=get_current_transformer_cfg(T, config).spark_config,
        )
        ...

    @abc.abstractmethod
    async def process_transformer_retry_hook(self) -> None:
        """A hook to call on failure of the process_transformer.

        Should be used with run_and_retry utility function.
        """
        ...



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import datetime
import os.path
import pathlib

from typing import Dict, List, Optional

import attrs
import pyspark.sql.functions as F
import pyspark.sql.types as T

from loguru import logger
from pyhocon import ConfigTree  # type: ignore
from pyspark.sql import DataFrame, DataFrameReader
from pyspark.sql.utils import AnalysisException

from pramen_py.metastore.reader_base import MetastoreReaderBase
from pramen_py.models import InfoDateSettings, MetastoreTable, TableFormat
from pramen_py.models.utils import (
    get_metastore_table,
    get_table_format_by_value,
)
from pramen_py.utils import convert_date_to_str, convert_str_to_date


@attrs.define(auto_attribs=True, slots=True)
class MetastoreReader(MetastoreReaderBase):
    """Interact with the data inside the file system in a convenient way.

    Provides interfaces for checking the data availability and
    retrieving these data.

    All methods which requires a table_name: str as input relies on
    the configuration. If the table_name will be missing in the config
    a KeyError will be raised.
    """

    HOCON_METASTORE_TABLES_VARIABLE = "pramen.metastore.tables"

    def get_latest(
        self,
        table_name: str,
        until: Optional[datetime.date] = None,
        uppercase_columns: bool = False,
    ) -> DataFrame:
        """Provides the latest partition of the table.

        if until is provided, then the partition will have until as
        upper boundary, otherwise the upper boundary will be the info_date
        """

        until = until or self.info_date
        metastore_table = get_metastore_table(table_name, self.tables)
        logger.info(
            f"Getting latest partition '{metastore_table.info_date_settings.column}'"
            f" for the table: '{table_name}' until {until}"
        )

        latest_date = convert_date_to_str(
            self.get_latest_available_date(table_name, until),
            fmt=metastore_table.info_date_settings.format,
        )

        if metastore_table.format == TableFormat.parquet:
            path = os.path.join(
                metastore_table.path,
                f"{metastore_table.info_date_settings.column}={latest_date}",
            )
            df = (
                self._get_dataframe_reader(
                    metastore_table.format, metastore_table.reader_options
                )
                .load(pathlib.Path(path).as_posix())
                .withColumn(
                    metastore_table.info_date_settings.column,
                    F.lit(latest_date).cast(T.DateType()),
                )
            )
        else:
            df = self._read_delta_table(metastore_table).filter(
                F.col(metastore_table.info_date_settings.column)
                == F.lit(latest_date)
            )

        logger.info(
            f"Table '{table_name}' successfully loaded for {latest_date} from"
            f" {metastore_table.path or metastore_table.table}"
        )
        return self._apply_uppercase_to_columns_names(df, uppercase_columns)

    def get_table(
        self,
        table_name: str,
        info_date_from: Optional[datetime.date] = None,
        info_date_to: Optional[datetime.date] = None,
        uppercase_columns: bool = False,
    ) -> DataFrame:
        """Get the table based on its name and config attributes.

        :param table_name:
        :param info_date_from: optional param with info_date as default
        :param info_date_to: optional param with info_date as default
        :param uppercase_columns: returns a table with uppercase column names

        if info_date_* params are provided, the data will be filtered
            based on it
        The data format (and other options) are obtained from the config.
        """

        metastore_table = get_metastore_table(table_name, self.tables)

        info_date_from_str = convert_date_to_str(
            info_date_from or self.info_date,
            fmt=metastore_table.info_date_settings.format,
        )
        info_date_to_str = convert_date_to_str(
            info_date_to or self.info_date,
            fmt=metastore_table.info_date_settings.format,
        )

        logger.info(
            f"Looking for {table_name} in the metastore where\n"
            f"info_date in range: {info_date_from_str} - {info_date_to_str}."
        )

        if metastore_table.format == TableFormat.parquet:
            df = self._read_parquet_table(metastore_table)
        else:
            df = self._read_delta_table(metastore_table)

        df_filtered = df.filter(
            F.col(metastore_table.info_date_settings.column)
            >= F.lit(info_date_from_str),
        ).filter(
            F.col(metastore_table.info_date_settings.column)
            <= F.lit(info_date_to_str),
        )

        logger.info(
            f"Table {table_name} successfully loaded from {metastore_table.path}."
        )

        return self._apply_uppercase_to_columns_names(
            df_filtered, uppercase_columns
        )

    def get_latest_available_date(
        self,
        table_name: str,
        until: Optional[datetime.date] = None,
    ) -> datetime.date:
        """Provide the latest info_date available for the table.

        if until is not specified then info_date is used.
        """

        until = until or self.info_date
        metastore_table = get_metastore_table(table_name, self.tables)

        if metastore_table.format == TableFormat.parquet:
            file_paths = self.fs_utils.list_files(
                metastore_table.path,
                file_pattern=f"{metastore_table.info_date_settings.column}=*",
            )
            info_dates = self._extract_info_dates_from_partition_paths(
                file_paths, metastore_table
            )
        else:
            dataframe = self._read_delta_table(metastore_table)
            info_dates = self._extract_info_dates_from_dataframe(
                dataframe, metastore_table
            )

        filtered_info_dates = self._filter_info_dates_by_until(
            info_dates, until
        )

        if not filtered_info_dates:
            info_dates_as_string = list(
                map(
                    lambda date: convert_date_to_str(
                        date, metastore_table.info_date_settings.format
                    ),
                    filtered_info_dates,
                )
            )
            raise ValueError(
                f"No partitions are available for the given '{metastore_table.name}'.\n"
                f"The table is available for the following dates:\n"
                f"{info_dates_as_string}\n"
                f"Only partitions earlier than {str(until)} might be included"
            )

        return max(filtered_info_dates)

    def is_data_available(
        self,
        table_name: str,
        from_date: Optional[datetime.date] = None,
        until_date: Optional[datetime.date] = None,
    ) -> bool:
        """Check if data is available for the given table.

        :param table_name:
        :param from_date: optional lower boundary. It is equal to the
            info_date by default
        :param until_date: optional upper boundary. It is equal to the
            info_date by default
        """

        from_date = from_date or self.info_date
        until_date = until_date or self.info_date

        logger.info(
            f"Checking for available data for {table_name} in period "
            f"between {from_date} to {until_date}"
        )

        try:
            latest_date = self.get_latest_available_date(
                table_name,
                until=until_date,
            )
        except ValueError:
            return False
        else:
            logger.debug(
                f"Data is available, checking if they are within necessary "
                f"date range: {from_date} - {until_date}"
            )
            return from_date <= latest_date <= until_date

    def from_config(self, config: ConfigTree):  # type: ignore
        tables = config.get(self.HOCON_METASTORE_TABLES_VARIABLE)
        metastore_tables = []
        for table in tables:
            metastore = MetastoreTable(
                name=table.get("name", ""),
                format=get_table_format_by_value(table.get("format", "")),
                path=table.get("path", ""),
                table=table.get("table", ""),
                description=table.get("description", ""),
                records_per_partition=table.get(
                    "records_per_partition", 500000
                ),
                info_date_settings=InfoDateSettings(
                    column=table.get(
                        "information.date.column",
                        config.get(
                            "pramen.information.date.column",
                            "pramen_info_date",
                        ),
                    ),
                    format=table.get("information.date.format", "yyyy-MM-dd"),
                    start=table.get("information.date.start", None),
                ),
                reader_options=table.get("read.option", {}),
                writer_options=table.get("write.option", {}),
            )
            metastore_tables.append(metastore)
        self.tables = metastore_tables.copy()
        return self

    @staticmethod
    def _apply_uppercase_to_columns_names(
        df: DataFrame, uppercase_columns: bool
    ) -> DataFrame:
        if uppercase_columns:
            return df.select([F.col(c).alias(c.upper()) for c in df.columns])
        else:
            return df

    @staticmethod
    def _filter_info_dates_by_until(
        dates: List[datetime.date],
        until: datetime.date,
    ) -> List[datetime.date]:
        def before_until(date: datetime.date) -> bool:
            return date <= until

        return list(filter(before_until, dates))

    @staticmethod
    def _extract_date_from_path(
        path: str, target_partition_name: str, date_format: str
    ) -> Optional[datetime.date]:
        def is_partition_hidden(column_name: str, column_value: str) -> bool:
            return column_name.startswith(
                ("_", ".")
            ) or column_value.startswith(("_", "."))

        file_name = path.replace("\\", "/").rsplit("/", 1)[-1]
        if not file_name:
            return None

        name_parts = file_name.split("=")
        if not len(name_parts) == 2:
            return None

        partition_name, partition_value = name_parts
        if is_partition_hidden(partition_name, partition_value):
            return None

        if target_partition_name != partition_name:
            raise ValueError(
                f"Partition name mismatch for path: {path}\n"
                f"Expected '{target_partition_name}' instead '{partition_name}'"
            )

        try:
            return convert_str_to_date(partition_value, fmt=date_format)
        except (KeyError, ValueError):
            raise ValueError(
                f"Date format mismatch for path: {path}"
                f"Expected '{date_format}' instead '{partition_value}'"
            )

    def _extract_info_dates_from_partition_paths(
        self,
        file_paths: List[str],
        table: MetastoreTable,
    ) -> List[datetime.date]:
        raw_info_dates = [
            self._extract_date_from_path(
                path,
                table.info_date_settings.column,
                table.info_date_settings.format,
            )
            for path in file_paths
        ]

        filtered_info_dates: List[datetime.date] = [
            info_date for info_date in raw_info_dates if info_date is not None
        ]

        if not filtered_info_dates:
            raise ValueError(
                f"The directory does not contain partitions by "
                f"'{table.info_date_settings.column}': {table.path}"
            )

        return filtered_info_dates

    @staticmethod
    def _extract_info_dates_from_dataframe(
        df: DataFrame, table: MetastoreTable
    ) -> List[datetime.date]:
        info_date_settings = table.info_date_settings
        info_dates_df = df.select(info_date_settings.column).distinct()
        distinct_info_dates = [row[0] for row in info_dates_df.collect()]

        if not distinct_info_dates:
            raise ValueError(
                f"Metastore table '{table.name}' at '{table.path or table.table}'"
                f" does not contain any info dates (column: {info_date_settings.column})"
            )

        return distinct_info_dates

    def _get_dataframe_reader(
        self, table_format: TableFormat, reader_options: Dict[str, str]
    ) -> DataFrameReader:
        return self.spark.read.options(**reader_options).format(
            table_format.value
        )

    def _read_parquet_table(
        self, metastore_table: MetastoreTable
    ) -> DataFrame:
        try:
            return self._get_dataframe_reader(
                TableFormat.parquet, metastore_table.reader_options
            ).load(metastore_table.path)
        except AnalysisException:
            raise Exception(
                f"Unable to access directory: {metastore_table.path}"
            )

    def _read_delta_table(self, metastore_table: MetastoreTable) -> DataFrame:
        reader = self._get_dataframe_reader(
            TableFormat.delta, metastore_table.reader_options
        )

        if metastore_table.path:
            return reader.load(metastore_table.path)
        elif metastore_table.table:
            return reader.table(metastore_table.table)
        else:
            raise ValueError(
                f"Metastore table '{metastore_table.name}' needs to contain either path or table option"
            )



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import abc
import datetime

from typing import List, Optional

import attrs

from pyspark.sql import DataFrame, SparkSession

from pramen_py.models import MetastoreTable
from pramen_py.utils.file_system import FileSystemUtils


@attrs.define(auto_attribs=True, slots=True)
class MetastoreReaderBase(metaclass=abc.ABCMeta):
    """Interact with the data inside the file system in a convenient way.

    All implementations of the MetastoreReader should be inherited
    from this base class

    Provides interfaces for checking the data availability and
    retrieving these data.

    All methods which requires a table_name: str as input relies on
    the configuration. If the table_name will be missing in the config
    a KeyError should be raised.
    """

    spark: SparkSession = attrs.field()
    tables: List[MetastoreTable] = attrs.field()
    info_date: datetime.date = attrs.field()
    fs_utils: FileSystemUtils = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        self.fs_utils = FileSystemUtils(spark=self.spark)

    @tables.default
    def _default_tables(self) -> List[MetastoreTable]:
        return []

    @info_date.default
    def _default_info_date(self) -> datetime.date:
        return datetime.date.today()

    def _set_tables(self, tables: List[MetastoreTable]) -> None:
        self.tables = tables

    def _set_info_date(self, info_date: datetime.date) -> None:
        self.info_date = info_date

    @tables.validator
    def _check_tables(
        self,
        _: attrs.Attribute,  # type: ignore
        value: List[MetastoreTable],
    ) -> None:
        for table in value:
            assert isinstance(table, MetastoreTable)

    @abc.abstractmethod
    def get_table(
        self,
        table_name: str,
        info_date_from: Optional[datetime.date] = None,
        info_date_to: Optional[datetime.date] = None,
        uppercase_columns: bool = False,
    ) -> DataFrame:
        """Get the table based on its name and config attributes.

        :param info_date_from: optional param with info_date as default
        :param info_date_to: optional param with info_date as default
        :param uppercase_columns: returns a table with uppercase column names

        if info_date_* params are provided, the data will be filtered
            based on it
        The data format (and other options) are obtained from the config.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_latest(
        self,
        table_name: str,
        until: Optional[datetime.date] = None,
        uppercase_columns: bool = False,
    ) -> DataFrame:
        """Provides the latest partition of the table.

        if until is provided, then the partition will have until as
        upper boundary, otherwise the upper boundary will be the info_date

        :param uppercase_columns: if True then all columns will be uppercase
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_latest_available_date(
        self,
        table_name: str,
        until: Optional[datetime.date] = None,
    ) -> datetime.date:
        """Provide the latest info_date available for the table.

        if until is not specified then info_date is used.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def is_data_available(
        self,
        table_name: str,
        from_date: Optional[datetime.date] = None,
        until_date: Optional[datetime.date] = None,
    ) -> bool:
        """Check if data is available for the given table.

        :param from_date: optional lower boundary. It is equal to the
            info_date by default
        :param until_date: optional upper boundary. It is equal to the
            info_date by default
        """

        raise NotImplementedError
